
/**
 * Title:        Appia<p>
 * Description:  Protocol development and composition framework<p>
 * Copyright:    Copyright (c) Alexandre Pinto & Hugo Miranda & Luis Rodrigues<p>
 * Company:      F.C.U.L.<p>
 * @author Alexandre Pinto & Hugo Miranda & Luis Rodrigues
 * @version 1.0
 */
package appia;

import appia.events.channel.*;

public class TimerManager implements Runnable {

	private class MyTimer {
		public String id;
		public long time;
		public long period;
		public ChannelEvent event;
		public MyTimer next;

		public MyTimer(String timerID, long time, long period, ChannelEvent event) {
			this.id=timerID;
			this.time=time;
			this.period=period;
			this.event=event;
		}
	}

	private boolean alive=false;
	private Thread thread;
	private MyTimer next=null;

	private class MyClock {
    public final long SYNC_TIME=1000;

		private long last_sync=0;
		private long now=0;

		public MyClock() {
			sync();
		}

		public synchronized void sync() {
			now=System.currentTimeMillis();
			last_sync=now;
		}

		public synchronized void update(long elapsed) {
			now+=elapsed;
		}

		public synchronized long read() {
			return now;
		}

		public synchronized void conditionalSync() {
			if (now > last_sync + SYNC_TIME) {
         now=System.currentTimeMillis();
         last_sync=now;
      }
		}
	}

	private MyClock clock=new MyClock();

	private synchronized void insert(MyTimer timer) {
		MyTimer prev=null;
		MyTimer t=next;

		while ( (t != null) && (t.time <= timer.time) ) {
			prev=t;
			t=t.next;
		}

		if (prev == null) {
			timer.next=next;
			next=timer;
		} else {
			timer.next=t;
			prev.next=timer;
		}
	}

	private synchronized void remove(String timerID) {
		MyTimer prev=null;
		MyTimer t=next;

		while ( (t != null) && (t.id != timerID) ) {
			prev=t;
			t=t.next;
		}

		if (t != null) {
			if (prev == null) {
				next=t.next;
			} else {
				prev.next=t.next;
			}
		}
	}

  private synchronized MyTimer getNextTimer(long now) {
    if ( (next != null) && (next.time <= now) ) {
      MyTimer t=next;
      next=next.next;
      return t;
    } else
      return null;
  }

  private synchronized void goToSleep(long now) {
    long sleep;
    if ( next != null ) {
      sleep=next.time-now;
    } else {
      sleep=clock.SYNC_TIME;
    }

    if ( sleep > 0 ) {
      try {
        this.wait(sleep);
        clock.update(sleep);
      } catch (InterruptedException e) {
        clock.sync();
      }
    }
  }

	private synchronized void setAlive(boolean alive) {
		this.alive=alive;
	}

	private synchronized boolean isAlive() {
		return alive;
	}


	public TimerManager() {
		thread=new Thread(this);
	}

	//////////////////////////////////////////////

	public long nextTimerEvent() { return -1; }

  public void handleTimerRequest(Timer timer) {
    EventQualifier q=timer.getQualifier();

    if ( !q.isNotify() ) {

      if ( q.isOn() ) {
        insert(new MyTimer(timer.timerID,timer.getTimeout(),0,timer));
      } else {
        remove(timer.timerID);
      }

      thread.interrupt();
    }
  }

	public void consumeTimerEvent() {}

  public void handlePeriodicTimer(PeriodicTimer timer) {
    EventQualifier q=timer.getQualifier();

    if ( !q.isNotify() ) {

      if ( q.isOn() ) {
        long period=timer.getPeriod();
        clock.sync();
        insert(new MyTimer(timer.timerID,clock.read()+period,period,timer));
      } else {
        remove(timer.timerID);
      }

      thread.interrupt();
    }
  }

	public void start() {
		setAlive(true);
		thread.start();
	}

	public void stop() {
		setAlive(false);
		thread.interrupt();
	}

	////////////////////////////////////////////

	public void run() {

		while (isAlive()) {
			clock.conditionalSync();

      long now=clock.read();
      MyTimer timer;

      while ( (timer=getNextTimer(now)) != null ) {

        ChannelEvent event;

        try {
          if (timer.period > 0) {
            event=(PeriodicTimer)timer.event.cloneEvent();
          } else {
            event=timer.event;
          }

          event.setSource(null);
          event.getDirection().direction=Direction.UP;
          event.getQualifier().set(EventQualifier.NOTIFY);

          event.init();
          event.getChannel().insertEvent(event);

          if (timer.period > 0) {
            timer.time=now+timer.period;
			      insert(timer);
          }
        }
        catch (AppiaException e) {
          throw new AppiaError("TimerManager:"+e.getMessage());
        }
        catch (CloneNotSupportedException e) {
          throw new AppiaError("TimerManager: CloneNotSupportedException ");
        }
      }

      goToSleep(now);
		}
	}
}