package appia.protocols.frag;

import appia.*;
import appia.events.*;
import appia.events.channel.*;
import appia.message.*;

import java.io.PrintStream;
import java.util.Enumeration;

/* Notes:
   messages are splited in the following way:
   - all but the last fragment are sent on FragEvent events
   - the last fragment is sent with the original event type
   
   the previous ones are all concatenated in the last one,
   and this one is the event sent to channel. 
   
   2 headers are appended:
   - a unique msgid
   - a fragment number
   
   incomplete messages are discarded some
   configurable time after the receipt of the 
   last fragment.
*/
 
public class FragSession extends Session {

    private int msgSeq=0;
    private java.util.Hashtable pdus;
    private java.util.Vector msgHolding;

    private final int fragHeaderSize=8;

//    private final boolean debugOn=false;

    private PrintStream debugOutput=null;

    public FragSession(Layer layer) {
	super(layer);
	pdus=new java.util.Hashtable(19);
	msgHolding=new java.util.Vector();
    }

    private void seqToByte(MsgBuffer buf,int seq) {
	long factor=(long)Math.pow(2,21);
	int off=buf.off+3;

	for(int i=3;i>=0;i--,factor/=128,off--) {
	    buf.data[off]=(byte)(seq/factor);
	    seq-=buf.data[off]*factor;
	}
    }

    private int byteToSeq(MsgBuffer buf) {
	int acum=0,factor=1,off=buf.off;

	for(int i=0;i<=3;i++,factor*=128,off++) {
	    acum+=buf.data[off]*factor;
	}

	return acum;
    }

    private FragHolder findFragHolder(Object who) {

	FragHolder f=null;
	java.util.Enumeration e=msgHolding.elements();

	while(e.hasMoreElements() && f==null) {
	    FragHolder t=(FragHolder)e.nextElement();
	    if(((appia.protocols.udpsimple.InetWithPort)t.source).equals((appia.protocols.udpsimple.InetWithPort)who)) f=t;
	}
	return f;
    }

    public void handle(Event e) {
	/* Main switch function */
	if(e instanceof Debug)
	    handleDebug((Debug)e);
	if(e instanceof ChannelInit)
	    queryPDUSize((ChannelInit)e);
	if(e instanceof MaxPDUSizeEvent)
	    setPDUSize((MaxPDUSizeEvent)e);
	if(e instanceof SendableEvent) {
	    if(e.getDirection().direction==Direction.UP)
		reassembly((SendableEvent)e);
	    else
		split((SendableEvent)e);
	}
	else {
	    /* For performance measures only */
	    try {
		e.go();
	    }
	    catch(AppiaEventException ex) {}
	}
    }

    private void queryPDUSize(ChannelInit e) {
	try{
	    /* Create the HashTable entry for this channel. The entry is
	       marked as undefined because we still don't know the max
	       pdu size
	    */
/*	    if(debugOn && debugOutput!=null)
		debugOutput.println("Frag: new channel opened. "+
				    "Sending query to retrieve PDU size.");
*/	    pdus.put(e.getChannel(),new PDUSize());
	    EchoEvent ee=new EchoEvent(new MaxPDUSizeEvent(),e.getChannel(),
				       new Direction(Direction.DOWN),this);
	    e.go();
	    ee.go();
	}
	catch(AppiaEventException ex) {
	    switch(ex.type) {
	    case ex.UNKNOWNSESSION :
		System.err.println("An exception stating that this "+
				   "class does not belong to the "+
				   "channel was raised in "+
				   "Frag");
		break;
	    case ex.ATTRIBUTEMISSING :
		System.err.println("Missing attribute exception in "+
				   "Frag");
		break;
	    case ex.UNWANTEDEVENT :
		System.err.println("The peer send an Unwanted event in "+
				   "Frag");
		break;
	    case ex.NOTINITIALIZED :
		System.err.println("Event not initialized exception "+
				   "ocurred after initialization in "+
				   "Frag");
		break;
	    }
	}
    }
    
    private void setPDUSize(MaxPDUSizeEvent e) {
	PDUSize p=(PDUSize)pdus.get(e.getChannel());
	p.size=e.pduSize-fragHeaderSize;
	p.def=true;
/*	if(debugOn && debugOutput!=null)
	    debugOutput.println("Frag: PDU size for channel "+
				e.getChannel().getChannelID()+" is "+p.size+
				" bytes");
*/	while(!p.holding.isEmpty()) {
/*	    if(debugOn && debugOutput!=null)
		debugOutput.println("Frag: forwarding messages waiting to "+
				    "know the PDU size");
*/	    split((SendableEvent)p.holding.removeFirst());
	}
	p.holding=null;
    }
    
    private void split(SendableEvent e) {
	PDUSize p=(PDUSize)pdus.get(e.getChannel());
	if(!p.def) {
/*	    if(debugOn && debugOutput!=null)
		debugOutput.println("Received message to unknown PDU size "+
				    "channel. Enqueing it.");
*/	    p.holding.add(e);
	}
	else 
	    splitMessage(e,p);
    }

    private void splitMessage(SendableEvent e, PDUSize p) {
	
	Message orig=e.getMessage();

	try {
	    int maxLength=p.size;
	    int fragNumber=0;
/*	    if(debugOn && debugOutput!=null)
		debugOutput.println("Frag: Message with "+orig.length()+
				    " bytes received for channel " +
				    e.getChannel().getChannelID() +
				    " (" + maxLength + " bytes max)");
*/	    while(orig.length()>maxLength) {
		FragEvent f=new FragEvent(e,this);
		Message m=new Message();
		orig.frag(m,maxLength);
/*		if(debugOn && debugOutput!=null)
		    debugOutput.println("Frag: orig with "+orig.length()+
					" m with "+m.length());
*/		MsgBuffer mbuf=new MsgBuffer();
		mbuf.len=fragHeaderSize;
		orig.push(mbuf);
/*		if(debugOn && debugOutput!=null)
		    debugOutput.println("Frag: mbuf length:"+mbuf.len+
					" mbuf off:"+mbuf.off);
*/
		seqToByte(mbuf,fragNumber++);
		mbuf.off+=4;
		seqToByte(mbuf,msgSeq);
		f.setMessage(orig);
		f.go();
		orig=m;
	    }
	    MsgBuffer mbuf=new MsgBuffer();
	    mbuf.len=fragHeaderSize;
	    orig.push(mbuf);
	    seqToByte(mbuf,fragNumber);
	    mbuf.off+=4;
	    seqToByte(mbuf,msgSeq++);
	    e.setMessage(orig);
	    e.go();
/*	    if(debugOn && debugOutput!=null)
		debugOutput.println("Frag: the message was splited in "+
				    (fragNumber+1)+ " pieces.");
*/	}
	catch(AppiaEventException ex) {
	    System.err.println("Unexpected event exception while "+
			       "fragmenting message");
	}
    }

    private void reassembly(SendableEvent e) {

	try {
	    /* Extract headers */
	    MsgBuffer mbuf=new MsgBuffer();
	    mbuf.len=fragHeaderSize;
	    e.getMessage().pop(mbuf);
	    int fragNumber=byteToSeq(mbuf);

	    /* Easiest case: unfragmented messages */
	    if(!(e instanceof FragEvent) && fragNumber==0) {
/*		if(debugOn && debugOutput!=null)
		    debugOutput.println("Frag: unfragmented message "+
					"received and forwarded.");
*/		e.go();
		return;
	    }

	    /* Concat fragment */

	    /* find fragment holder */
	    FragHolder fHold=findFragHolder(e.source);

	    /* Drop the fragment if:
	       - it is a new message but the fragment received is
	         not the first
	       - the fragment is out of order
	    */
	    if(fHold==null && fragNumber!=0) {
/*		if(debugOn && debugOutput!=null) 
		    debugOutput.println("Frag: Fragment out of order ("+
					fragNumber+") from "+
					"unknown sequence received. "+
					"Discarded");
*/		return;
	    }

	    mbuf.off+=4;
	    int msgId=byteToSeq(mbuf);

	    if(fHold!=null && msgId!=fHold.msgId && fragNumber==0) {
/*		if(debugOn && debugOutput!=null)
		    debugOutput.println("Frag: Message out of order from "+
					"known sender. Discarding existing" +
					"fragments.");
*/		fHold.reset(e.getMessage(),msgId);
		return;
	    }
	    if(fHold!=null && (fragNumber!=fHold.next || msgId!=fHold.msgId)) {
/*		if(debugOn && debugOutput!=null) 
		    debugOutput.println("Frag: Fragment out of order from "+
					"known sequence received. "+
					"Discarded all the sequence.");
*/		msgHolding.remove(fHold);
		return;
	    }

	    /* Create new fHolder */
	    if(fHold==null) {
		fHold=new FragHolder(e.source,e.getMessage(),msgId);
		msgHolding.add(fHold);
/*		if(debugOn && debugOutput!=null)
		    debugOutput.println("Frag: creating new sequence for "+
					"channel "+
					e.getChannel().getChannelID());
*/	    }
	    else if(e instanceof FragEvent) {
		fHold.message.join(e.getMessage());
/*		if(debugOn && debugOutput!=null)
		    debugOutput.println("Frag: appending fragment "+
					fHold.next+" for sequence in "+
					"channel "+
					e.getChannel().getChannelID() +
					". Message actual size:"+
					fHold.message.length());
*/		fHold.next++;
	    }
	    else {
		/* last case: the last fragment is contained
		   in one event of the original type. Append the
		   remaining header and forward it
		*/
		fHold.message.join(e.getMessage());
		e.setMessage(fHold.message);
/*		if(debugOn && debugOutput!=null)
		    debugOutput.println("Frag: appending last fragment ("+
					fHold.next +") for sequence in "+
					"channel "+
					e.getChannel().getChannelID()+
					". Message final size:"+
					fHold.message.length());
*/		msgHolding.remove(fHold);
		e.go();
	    }
	}
	catch(AppiaEventException ex) {
	    System.err.println("Unexpected event exception while "+
			       "reassemblying message");
	}
    }

    private void handleDebug(Debug e) {

	EventQualifier q=e.getQualifier();

	if(q.isOn()) {
	    debugOutput=new PrintStream(e.getOutput());
	    debugOutput.println("Frag: Debugging started");
	}
	else if(q.isOff()) {
	    debugOutput=null;
	}
	else if(q.isNotify()) {
	    printState(new PrintStream(e.getOutput()));
	}

	try {
	    e.go();
	}
	catch(AppiaEventException ex) {
	}
    }

    private void printState(PrintStream out) {
	out.println("Frag Session state dumping:");

	Enumeration en=pdus.keys();

	if(en.hasMoreElements())
	    out.println("List of channels:");
	else
	    out.println("No channels available");

	while(en.hasMoreElements()) {
	    Channel c=(Channel)en.nextElement();
	    out.print("Channel name: "+c.getChannelID()+
		      " Defined: ");
	    PDUSize p=(PDUSize)pdus.get(c);
	    if(p.def)
		out.println("yes. Max PDU size: "+p.size);
	    else
		out.println("no.");
	}

	en=msgHolding.elements();

	if(en.hasMoreElements())
	    out.println("Pending messages:");
	else
	    out.println("No pending messages.");

	while(en.hasMoreElements()) {
	    FragHolder f=(FragHolder)en.nextElement();
	    out.println("Current message size: " + f.message.length() +
			" Fragments received: " + f.next);
	}

	out.println("Debug output is currently "+ 
		    (debugOutput==null ? "off" : "on"));
    }
}
    
