
/**
 * Title:        Apia<p>
 * Description:  Protocol development and composition framework<p>
 * Copyright:    Copyright (c) Alexandre Pinto & Hugo Miranda & Luis Rodrigues<p>
 * Company:      F.C.U.L.<p>
 * @author Alexandre Pinto & Hugo Miranda & Luis Rodrigues
 * @version 1.0
 */
package appia.protocols.group;

import appia.*;
import appia.protocols.group.*;
import appia.protocols.group.events.*;
import appia.events.channel.EchoEvent;
import appia.protocols.udpsimple.InetWithPort;

import appia.events.channel.Debug;
import java.io.PrintStream;

public class IntraSession extends Session {

  public static final int K=2;

  public IntraSession(Layer layer) {
    super(layer);
  }

  public void handle(Event event) {

    // InstallView
    if (event instanceof InstallView) {
      handleInstallView((InstallView)event);
      return;
    }

    // Fail
    if (event instanceof Fail) {
      handleFail((Fail)event);
      return;
    }

    // BlockOk
    if (event instanceof BlockOk) {
      if (event.getDirection().direction == Direction.UP) {
        handleUpBlockOk((BlockOk)event);
      } else {
        try { event.go(); } catch (AppiaEventException ex) { ex.printStackTrace(); }
      }
      return;
    }

    // Stable
    if (event instanceof Stable) {
      handleStable((Stable)event);
      return;
    }

    // GroupInit
    if (event instanceof GroupInit) {
      handleGroupInit((GroupInit)event);
      return;
    }

    // View
    if (event instanceof View) {
      handleView((View)event);
      return;
    }

    // Debug
    if (event instanceof Debug) {
      Debug ev=(Debug)event;

      if (ev.getQualifier().isOn()) {
        if (ev.getOutput() instanceof PrintStream)
           debug=(PrintStream)ev.getOutput();
        else
          debug=new PrintStream(ev.getOutput());
      } else {
        if (ev.getQualifier().isOff())
           debug=null;
      }
    }

    debug("event (\""+event.getClass().getName()+"\") discarded because it is unwanted");
  }

  private ViewState vs;
  private LocalState ls;
  private Endpt my_endpt;
  private boolean new_view;
  private boolean installing;
  private boolean blocked;
  private boolean stabilized;
  private boolean stable;

  private void handleGroupInit(GroupInit ev) {
    my_endpt=ev.vs.view[ev.rank];

    // values already tested by GroupInit
    ViewState new_vs=ev.vs;
    LocalState new_ls=new LocalState(new_vs,my_endpt);

    try { ev.go(); } catch (AppiaEventException ex) { ex.printStackTrace(); }

    sendView(new_vs,new_ls,ev.getChannel());
  }

  private void handleView(View ev) {

    vs=ev.vs;
    ls=ev.ls;

    new_view=false;
    installing=false;
    blocked=false;
    stabilized=false;
    stable=false;

    try { ev.go(); } catch (AppiaEventException ex) { ex.printStackTrace(); }
  }

  private void handleInstallView(InstallView ev) {

    // i am waiting for a View event
    if (installing)
       return;

    if (ls.failed[ev.orig]) {
      debug("discarded InstallView because it came from a failed member");
      return;
    }

    Integer k=(Integer)ev.getObjectsMessage().pop();

    ViewState new_vs=(ViewState)ev.getObjectsMessage().pop();

    // see if i am not a member of the new view
    if (new_vs.getRank(my_endpt) < 0) {
      debug("discarded InstallView because i don't belong to view");
      return;
    }

    LocalState new_ls=new LocalState(new_vs,my_endpt);

    // resend InstallView
    if (new_ls.my_rank < k.intValue()) {
      try {
        ev.getObjectsMessage().push(new_vs);
        ev.getObjectsMessage().push(k);
        ev.getDirection().direction=Direction.DOWN;
        ev.setSource(this);
        ev.init();
        ev.go();
      } catch (AppiaEventException ex) {
        ex.printStackTrace();
        debug("impossible to resend InstallView");
      }
    }

    sendView(new_vs,new_ls,ev.getChannel());
  }

  private void handleFail(Fail ev) {
    try { ev.go(); } catch (AppiaEventException ex) { ex.printStackTrace(); }

    if (ls.am_coord && !new_view)
       sendBlock(ev.getChannel());
  }

  private void handleUpBlockOk(BlockOk ev) {
    blocked=true;
    try {
      Stabilize stabilize=new Stabilize(ev.getChannel(),new Direction(Direction.DOWN),this,vs.group,vs.id);
      stabilize.go();
    } catch (AppiaEventException ex) {
      ex.printStackTrace();
      throw new AppiaGroupError("IntraSession: impossible to send Stabilize");
    }
  }

  private void handleStable(Stable ev) {
    stable=true;
    stabilized=ev.all;

    try { ev.go(); } catch (AppiaEventException ex) { ex.printStackTrace(); }

    if (ls.am_coord && stabilized) {

      int i,j,size;

      size=0;
      for(i=0 ; i < ls.failed.length ; i++) {
        if (!ls.failed[i])
           size++;
      }

      Endpt[] new_view=new Endpt[size];
      InetWithPort[] new_addrs=new InetWithPort[size];

      j=0;
      for(i=0 ; i < ls.failed.length ; i++) {
        if (!ls.failed[i]) {
          new_view[j]=vs.view[i];
          new_addrs[j]=vs.addresses[i];
          j++;
        }
      }

      ViewID new_id=vs.id.next(my_endpt);
      ViewID[] new_previous=new ViewID[1];
      new_previous[0]=vs.id;

      ViewState new_vs=null;
      LocalState new_ls=null;

      try {
        new_vs=new ViewState(vs.version,vs.group,new_id,new_previous,new_view,new_addrs);
        new_ls=new LocalState(new_vs,my_endpt);
      } catch (Exception ex) {
        ex.printStackTrace();
        throw new AppiaGroupError("IntraSession: impossible to create new view state");
      }

      sendInstallView(ev.getChannel(),new_vs,K);
      sendView(new_vs,new_ls,ev.getChannel());
    }
  }

  private void sendView(ViewState vs, LocalState ls, Channel channel) {
    installing=true;
    try {
      View view=new View(vs,ls);
      EchoEvent echo=new EchoEvent(view,channel,new Direction(Direction.DOWN),this);
      echo.go();
    } catch (AppiaEventException ex) {
      ex.printStackTrace();
      System.err.println("appia:group:IntraSession: impossible to send View");
    }
  }

  private void sendSuspect(int rank, Channel channel) {
    boolean[] suspects=new boolean[ls.failed.length];
    int i;
    for(i=0 ; i < suspects.length ; i++) suspects[i]= (i==rank);

    try {
      Suspect ev=new Suspect(suspects,channel,new Direction(Direction.DOWN),this,vs.group,vs.id);
      ev.go();
    } catch (AppiaEventException ex) {
      ex.printStackTrace();
      System.err.print("appia:group:IntraSession: impossible to send suspition");
    }
  }

  private void sendBlock(Channel channel) {
    new_view=true;
    try {
      Block ev=new Block(channel,new Direction(Direction.DOWN),this,vs.group,vs.id);
      ev.go();
    } catch (AppiaEventException ex) {
      ex.printStackTrace();
      throw new AppiaGroupError("IntraSession: impossible to send Block");
    }
  }


  private void sendInstallView(Channel channel, ViewState new_vs, int k) {
    try {
      InstallView ev=new InstallView(channel,new Direction(Direction.DOWN),this,vs.group,vs.id);
      ev.getObjectsMessage().push(new_vs);
      ev.getObjectsMessage().push(new Integer(k));
      ev.init();
      ev.go();
    } catch (AppiaEventException ex) {
      ex.printStackTrace();
      debug("impossible to send InstallView");
    }
    /*
    int i;
    for(i=0 ; i < vs.addresses.length ; i++) {
      if (i != my_rank) {
        try {
          InstallView ev=new InstallView(channel,new Direction(Direction.DOWN),this);
          ev.getObjectsMessage().push(vs);
          ev.getObjectsMessage().push(new Integer(k));
          ev.getObjectsMessage().flush();
          ev.dest=vs.addresses[i];
          ev.source=vs.addresses[my_rank];
          ev.go();
        } catch (AppiaEventException ex) {
          ex.printStackTrace();
          System.err.println("appia:group:IntraSession: impossible to send InstallView to rank "+i);
        } catch (java.io.IOException ex) {
          ex.printStackTrace();
          System.err.println("appia:group:IntraSession: impossible to send InstallView to rank "+i);
        }
      }
    }
    */
  }

  private PrintStream debug=null;

  private void debug(String s) {
    if (debug != null)
       debug.println("appia:group:IntraSession: "+s);
  }
}