
/**
 * Title:        Apia<p>
 * Description:  Protocol development and composition framework<p>
 * Copyright:    Copyright (c) Alexandre Pinto & Hugo Miranda & Luis Rodrigues<p>
 * Company:      F.C.U.L.<p>
 * @author Alexandre Pinto & Hugo Miranda & Luis Rodrigues
 * @version 1.0
 */
package appia.protocols.group;

import appia.*;
import appia.events.channel.PeriodicTimer;
import appia.protocols.group.*;
import appia.protocols.group.events.*;

import appia.events.channel.Debug;
import java.io.PrintStream;

public class StableSession extends Session {

  private long gossip_time;
  private long last_gossip;

  public StableSession(Layer layer, long gossip_time) {
    super(layer);
    this.gossip_time=gossip_time;
  }

  public void handle(Event event) {

    // StableGossip
    if (event instanceof StableGossip) {
      handleGossip((StableGossip)event);
      return;
    }

    // Stabilize
    if (event instanceof Stabilize) {
      handleStabilize((Stabilize)event);
      return;
    }

    // Retransmission
    if (event instanceof Retransmission) {
      handleRetransmission((Retransmission)event);
      return;
    }

    // BlockOk
    if (event instanceof BlockOk) {
      if (event.getDirection().direction == Direction.DOWN) {
        ((BlockOk)event).getObjectsMessage().push(table[ls.my_rank]);
      } else {
        handleGossip((BlockOk)event);
      }
      try { event.go(); } catch (AppiaEventException ex) { ex.printStackTrace(); }
      return;
    }

    // GroupSendableEvent
    // NOTE: it should allways be after the BlockOk, StableGossip, Stabilize and Retransmission
    //       to avoid overriding those handlers
    if (event instanceof GroupSendableEvent) {
      handleGroupSendableEvent((GroupSendableEvent)event);
      return;
    }

    // PeriodicTimer
    if (event instanceof PeriodicTimer) {
      PeriodicTimer ev=(PeriodicTimer)event;

      try { ev.go(); } catch (AppiaEventException ex) { ex.printStackTrace(); }

      if ((vs != null) && (ls != null) && ev.getQualifier().isNotify()) {
        long now=System.currentTimeMillis();

        if (now-last_gossip >= gossip_time) {
          gossip(event.getChannel());
          last_gossip=now;
        }
      }

      return;
    }

    // Fail
    if (event instanceof Fail) {
      handleFail((Fail)event);
      try { event.go(); } catch (AppiaEventException ex) { ex.printStackTrace(); }
      return;
    }

    // View
    if (event instanceof View) {
      View ev=(View)event;

      vs=ev.vs;
      ls=ev.ls;

      try { ev.go(); } catch (AppiaEventException ex) { ex.printStackTrace(); }

      init();

      return;
    }

    if (event instanceof Debug) {
      Debug ev=(Debug)event;

      if (ev.getQualifier().isOn()) {
        if (ev.getOutput() instanceof PrintStream)
           debug=(PrintStream)ev.getOutput();
        else
          debug=new PrintStream(ev.getOutput());
      } else {
        if (ev.getQualifier().isOff())
           debug=null;
      }
    }

    debug("discarded unwanted event \""+event.getClass().getName()+"\"");
  }

  private ViewState vs;
  private LocalState ls;
  private long[][] table;
  private boolean stabilizing;
  private boolean stabilize;
  private StableStorage storage=new StableStorage(0);

  private void init() {
    table=new long[vs.view.length][vs.view.length];
    stabilizing=false;
    stabilize=false;
    storage.reset(vs.view.length);
    retransmit_max=null;
    retransmit_rank=null;
  }

  // the parameter is GroupSendableEvent and not StableGossip because BlockOk also carries
  // a gossip message
  private void handleGossip(GroupSendableEvent ev) {

    long[] row=(long[])ev.getObjectsMessage().pop();

    int col;
    for (col=0 ; col < table[ev.orig].length ; col++) {
      if (row[col] > table[ev.orig][col])
         table[ev.orig][col]=row[col];
    }

    clean_storage();

    if (stabilizing && stabilize)
       stable(ev.getChannel());
  }

  private void handleStabilize(Stabilize ev) {
    stabilizing=true;

    if (ev.getDirection().direction == Direction.DOWN) {
      stabilize=true;

      ev.getObjectsMessage().push(generateRetransmissions(ev.getChannel()));

      try { ev.go(); } catch (AppiaEventException ex) { ex.printStackTrace(); }

    } else {
      Retransmit r=(Retransmit)ev.getObjectsMessage().pop();

      Channel c=ev.getChannel();

      while (r != null) {
        if (r.retransmiter == ls.my_rank)
           retransmit(r.dest,r.orig,r.lo,r.hi,c);
        r=r.next;
      }
    }

    stable(ev.getChannel());
  }

  private void handleGroupSendableEvent(GroupSendableEvent ev) {

    if (ev.getDirection().direction == Direction.DOWN) {
      Long seqno=new Long(table[ls.my_rank][ls.my_rank]+1);

      ev.getObjectsMessage().push(seqno);

      try {
        ev.go();
        table[ls.my_rank][ls.my_rank]++;
      } catch (AppiaEventException ex) {
        ex.printStackTrace();
        System.err.println("appia:group:StableSession: down event discarded");
      }
    } else { // UP

      long seqno=((Long)ev.getObjectsMessage().pop()).longValue();

      if (seqno != table[ls.my_rank][ev.orig]+1) {
        System.err.println("appia:group:StableSession: up event discarded due to wrong seq. number (had "+table[ls.my_rank][ev.orig]+", received "+seqno+")");
        return;
      }

      table[ls.my_rank][ev.orig]=seqno;

      storage.add(ev.orig,seqno,ev);

      if (stabilizing)
         stable(ev.getChannel());

      try {
        ev.go();
      } catch (AppiaEventException ex) {
        ex.printStackTrace();
        System.err.println("appia:group:StableSession: up event discarded");
      }
    }
  }

  private void handleRetransmission(Retransmission ev) {
    int orig=((Integer)ev.getObjectsMessage().pop()).intValue();
    String event=(String)ev.getObjectsMessage().pop();

    try {
      GroupSendableEvent retrans=(GroupSendableEvent)Class.forName(event).newInstance();
      retrans.setObjectsMessage(ev.getObjectsMessage());
      retrans.source=vs.view[orig];
      retrans.orig=orig;
      retrans.setChannel(ev.getChannel());
      retrans.setDirection(ev.getDirection());
      retrans.setSource(this);

      retrans.init();
      handleGroupSendableEvent(retrans);

      return;

    } catch (ClassNotFoundException ex) { ex.printStackTrace(); }
      catch (IllegalAccessException ex) { ex.printStackTrace(); }
      catch (InstantiationException ex) { ex.printStackTrace(); }
      catch (AppiaEventException ex) { ex.printStackTrace(); }

    System.err.println("appia:group:StableSession: retransmission discarded");
  }

  private void handleFail(Fail ev) {

    if (stabilize) {
      Retransmit r=generateRetransmissions(ev.getChannel());

      if (r != null) {
        try {
          Stabilize stabilize=new Stabilize(ev.getChannel(),new Direction(Direction.DOWN),this,vs.group,vs.id);
          stabilize.getObjectsMessage().push(r);
          stabilize.go();
        } catch (AppiaEventException ex) {
          ex.printStackTrace();
          System.err.println("appia:group:StableSession: impossible to send new Stabilize");
        }
      }
    }

    if (stabilizing && stabilize)
       stable(ev.getChannel());
  }

  private void gossip(Channel channel) {
    try {
      StableGossip ev=new StableGossip(channel,new Direction(Direction.DOWN),this,vs.group,vs.id);
      ev.getObjectsMessage().push(table[ls.my_rank]);

      ev.go();
    } catch (AppiaEventException ex) {
      ex.printStackTrace();
      System.err.println("appia:group:StableSession: impossible to send gossip");
    }
  }

  private void stable(Channel channel) {
    boolean stable=true;

    int row,col;

    if (table.length > 1) {
      if (stabilize) {
        // every one is stable
        for (row=0 ; (row < table.length) && stable ; row++) {
          if (!ls.failed[row]) {
            for (col=0 ; (col < table[row].length) && stable ; col++) {
              stable= table[row][col]==table[ls.coord][col];
            }
          }
        }
      } else {
        // only i am stable
        for (col=0 ; (col < table[0].length) && stable ; col++) {
          long max=-1;

          for(row=ls.coord ; (row < table.length) && stable ; row++) {
            if (table[row][col] > max)
               max=table[row][col];
          }

          stable= table[ls.my_rank][col]==max;
        }
      }
    }

    if (stable) {
      try {
        Stable ev=new Stable(table[ls.my_rank],stabilize,channel,new Direction(Direction.UP),this,vs.group,vs.id);
        ev.go();
      } catch (AppiaEventException ex) {
        ex.printStackTrace();
        System.err.println("appia:group:StableSession: impossible to send stable indication");
      }
    }
  }

  private void clean_storage() {
    int row,col;
    long min;

    for (col=0 ; col < vs.view.length ; col++) {
      if (col != ls.my_rank) {
        min=Integer.MAX_VALUE;

        for (row=0 ; row < table.length ; row++) {
          if (!ls.failed[row] && (table[row][col] < min))
             min=table[row][col];
        }

        if (min >= storage.getFirstSeqno(col))
           storage.removeUntil(col,min);
      }
    }
  }

  // Retransmissions

  private long[] retransmit_max=null;
  private int[] retransmit_rank=null;

  private Retransmit generateRetransmissions(Channel channel) {
    Retransmit r=null;
    int col;

    if (retransmit_max == null) {
      retransmit_max=new long[table.length];
      retransmit_rank=new int[table.length];

      for (col=0 ; col < table.length ; col++) {
        origMax(col);
        r=origRetransmissions(col,r,channel);
      }

    } else {

      for(col=0 ; col < retransmit_rank.length ; col++) {
        if (ls.failed[retransmit_rank[col]]) {
          origMax(col);
          r=origRetransmissions(col,r,channel);
        }
      }
    }

    return r;
  }

  private void origMax(int col) {
    retransmit_max[col]=-1;
    retransmit_rank[col]=-1;

    int row;

    for(row=0 ; row < table.length ; row++) {
      //TODO: remove the condition "row != col". for testing porpuses only.
      //if ((row != col) && !ls.failed[row] && (table[row][col] > retransmit_max[col])) {
      if (!ls.failed[row] && (table[row][col] > retransmit_max[col])) {
        retransmit_max[col]=table[row][col];
        retransmit_rank[col]=row;
      }
    }
  }

  private Retransmit origRetransmissions(int col, Retransmit r, Channel channel) {

    //The following test propably shouldn't be necessary. The problem is that i don't
    // keep here my sent messages, this is done by the FIFO layer. Therefore no one can
    // retransmit their one messages. I expect FIFO will do so in the very near future
    if (retransmit_rank[col] == col)
       return r;

    int row;

    for(row=0 ; row < table.length ; row++) {
      if (!ls.failed[row] && (table[row][col] < retransmit_max[col])) {
        if (ls.my_rank == retransmit_rank[col]) {
          retransmit(row,col,table[row][col],retransmit_max[col],channel);
          if (debug != null) debug("retransmiting myself to "+row+" from "+col+" ("+table[row][col]+"->"+retransmit_max[col]+")");
        } else {
          r=new Retransmit(retransmit_rank[col],row,col,table[row][col],retransmit_max[col],r);
          if (debug != null) debug("ordering retransmission by "+retransmit_rank[col]+" to "+row+" from "+col+" ("+table[row][col]+"->"+retransmit_max[col]+")");
        }
      }
    }

    return r;
  }

  private void retransmit(int dest, int orig, long lo, long hi, Channel channel) {
    long seqno=Math.max(lo,storage.getFirstSeqno(orig));
    Retransmission ev;

    int[] dests={dest};

    while (seqno <= hi) {
      ev=storage.getRetransmission(orig,seqno);

      if (ev == null)
         throw new AppiaGroupError("StableSession: unknown seqno !!!!!");

      ev.dest=dests;
      ev.group=vs.group;
      ev.view_id=vs.id;
      ev.setChannel(channel);
      ev.setDirection(new Direction(Direction.DOWN));
      ev.setSource(this);

      try {
        ev.init();
        ev.go();
      } catch (AppiaEventException ex) {
        ex.printStackTrace();
        System.err.println("appia:group:StableSession: impossible to create retransmission");
      }

      seqno++;
    }
  }

  private static class Retransmit implements java.io.Serializable {

    public int retransmiter;
    public int dest;
    public int orig;
    public long lo;
    public long hi;
    public Retransmit next;

    public Retransmit(int r, int d, int o, long l, long h, Retransmit n) {
      retransmiter=r;
      dest=d;
      orig=o;
      lo=l;
      hi=h;
      next=n;
    }
  }

  private PrintStream debug=null;

  private void debug(String s) {
    if (debug != null)
       debug.println("appia:group:StableSession: "+s);
  }
}