
/**
 * Title:        Apia<p>
 * Description:  Protocol development and composition framework<p>
 * Copyright:    Copyright (c) Alexandre Pinto & Hugo Miranda & Luis Rodrigues<p>
 * Company:      F.C.U.L.<p>
 * @author Alexandre Pinto & Hugo Miranda & Luis Rodrigues
 * @version 1.0
 */
package appia.protocols.group;

import appia.protocols.group.events.ObjectsMessage;
import appia.protocols.group.events.GroupSendableEvent;
import appia.protocols.group.events.Retransmission;

public class StableStorage {

  private class Entry {
    public long seqno;
    public String event;
    public ObjectsMessage omsg;

    public Entry next;
    public Entry prev;
  }

  private class Header {
    public Entry first;
    public Entry last;
  }

  private class EntryPool {

    public static final long MIN=100;

    Entry free=null;
    long used=0;
    long total=0;
    long min_threshold;

    public EntryPool() {
      increase();
    }

    public void increase() {
      long i;
      long newtotal= (total == 0) ? MIN : total*2 ;

      for(i=total ; i < newtotal ; i++) {
        Entry e=new Entry();
        if (free != null)
           free.prev=e;
        e.next=free;
        e.prev=null;
        free=e;
      }

      total=newtotal;
      min_threshold= (total <= MIN) ? (-1) : (total/4) ;
    }

    public void decrease() {
      if (total > MIN) {
        long i;
        long newtotal=total/2;

        for(i=total ; i > newtotal ; i--) {
          free=free.next;
        }

        if (free != null)
           free.prev=null;

        total=newtotal;
        min_threshold= (total <= MIN) ? (-1) : (total/4) ;
      }
    }

    public Entry get() {
      Entry e=free;
      free=free.next;
      if (free != null)
         free.prev=null;
      used++;

      if (used == total)
         increase();

      return e;
    }

    public void put(Entry e) {
      e.event=null;
      e.omsg=null;

      e.next=free;
      e.prev=null;
      if (free != null)
         free.prev=e;
      free=e;
      used--;

      if (used == min_threshold)
         decrease();
    }
  }

  private EntryPool pool=new EntryPool();
  private Header[] array;

  public StableStorage(int size) {
    array=new Header[size];
    int i;
    for (i=0 ; i < array.length ; i++) array[i]=new Header();
  }

  public void reset(int size) {
    int i;
    Entry e,aux;

    for (i=0 ; i < array.length ; i++) {
      e=array[i].first;
      while (e != null) {
        aux=e;
        e=e.next;
        pool.put(aux);
      }
    }

    array=new Header[size];
    for (i=0 ; i < array.length ; i++) array[i]=new Header();
  }

  public void add(int index, long seqno, GroupSendableEvent ev) {

    Header h=array[index];
    Entry e=h.last;
    Entry enew=pool.get();

    enew.seqno=seqno;
    enew.event=ev.getClass().getName();
    enew.omsg=(ObjectsMessage)ev.getObjectsMessage().clone();

    while ( (e != null) && (e.seqno > seqno) ) e=e.next;

    if (e != null) {
      enew.prev=e;
      enew.next=e.next;
      if (e.next != null)
         e.next.prev=enew;
      else
        h.last=enew;
      e.next=enew;
    } else {
      enew.prev=null;
      enew.next=h.first;
      if (h.first != null)
         h.first.prev=enew;
      else
        h.last=enew;
      h.first=enew;
    }
  }

  public Retransmission getRetransmission(int index, long seqno) {

    Entry e=array[index].first;

    while ( (e != null) && (e.seqno != seqno) ) e=e.next;

    if (e == null)
       return null;

    Retransmission ev=new Retransmission();
    e.omsg.push(new Long(e.seqno));
    e.omsg.push(e.event);
    e.omsg.push(new Integer(index));
    ev.setObjectsMessage(e.omsg);
    return ev;
  }

  public void removeUntil(int index, long seqno) {
    Header h=array[index];
    Entry aux;

    while ( (h.first != null) && (h.first.seqno <= seqno) ) {
      aux=h.first;
      h.first=h.first.next;
      pool.put(aux);
    }

    if (h.first != null)
       h.first.prev=null;
    else
      h.last=null;
  }

  public long getLastSeqno(int index) {
    if (array[index].last != null)
       return array[index].last.seqno;
    else
      return -1;
  }

  public long getFirstSeqno(int index) {
    if (array[index].first != null)
       return array[index].first.seqno;
    else
      return -1;
  }
}