
/**
 * Title:        Apia<p>
 * Description:  Protocol development and composition framework<p>
 * Copyright:    Copyright (c) Alexandre Pinto & Hugo Miranda & Luis Rodrigues<p>
 * Company:      F.C.U.L.<p>
 * @author Alexandre Pinto & Hugo Miranda & Luis Rodrigues
 * @version 1.0
 */
package appia.protocols.group;

import appia.Session;

import appia.*;
import appia.protocols.group.*;
import appia.protocols.group.events.*;
import appia.events.channel.EchoEvent;

public class SuspectSession extends Session {

  private long suspect_sweep;
  private int suspect_idle;

  public SuspectSession(Layer layer, long suspect_sweep, int suspect_idle) {
    super(layer);
    this.suspect_sweep=suspect_sweep;
    this.suspect_idle=suspect_idle;
  }

  public void handle(Event event) {

    // Alive
    if (event instanceof Alive) {
      suspects[((Alive)event).orig]=0;
      return;
    }

    // Suspect
    if (event instanceof Suspect) {
      handleSuspect((Suspect)event);
      return;
    }

    // SuspectTimer
    if (event instanceof SuspectTimer) {
      handleSuspectTimer((SuspectTimer)event);
      try { event.go(); } catch (AppiaEventException ex) { ex.printStackTrace(); }
      return;
    }

    // View
    if (event instanceof View) {
      handleView((View)event);
      return;
    }

    System.err.println("appia:group:SuspectSession: received unwanted event (\""+event.getClass().getName()+"\")");
    try { event.go(); } catch (AppiaEventException ex) { ex.printStackTrace(); }
  }

  private ViewState vs;
  private LocalState ls;
  private int[] suspects=null;

  private void handleView(View ev) {
    vs=ev.vs;
    ls=ev.ls;

    try { ev.go(); } catch (AppiaEventException ex) { ex.printStackTrace(); }

    if (suspects == null) {
      try {
        SuspectTimer periodic=new SuspectTimer("Suspect Timer",suspect_sweep,ev.getChannel(),this);
        periodic.go();
      } catch (AppiaException ex) {
        ex.printStackTrace();
        System.err.println("appia:group:SuspectSession: impossible to set SuspectTimer, SuspectSession will be idle");
      }
    }

    suspects=new int[vs.view.length];
    for (int i=0 ; i < suspects.length ; i++) suspects[i]=0;
  }

  private void handleSuspect(Suspect ev) {

    if (ev.getDirection().direction == Direction.UP) {
      if ( ls.failed[ev.orig] ) {
        System.err.println("appia:group:SuspectSession: invalid (failed) message source");
        return;
      }

      ev.failed=(boolean[])ev.getObjectsMessage().pop();
    }

    if (ev.failed[ls.my_rank]) {
      System.err.println("appia:group:SuspectSession: i am not failed !!");
      return;
    }

    int i;
    boolean new_fail=false;
    boolean[] new_failed=null;

    for (i=0 ; i < ev.failed.length ; i++) {
      if (ev.failed[i] && !ls.failed[i]) {
        ls.fail(i);
        if (!new_fail) {
          new_fail=true;
          new_failed=new boolean[ls.failed.length];
        }
        new_failed[i]=true;
      }
    }

    if (new_fail) {
      if (ev.getDirection().direction == Direction.DOWN) {
        ev.getObjectsMessage().push(ls.failed);

        try {
          ev.go();
        } catch (AppiaEventException ex) {
          ex.printStackTrace();
          System.err.println("appia:group:SuspectSession: unable to inform others of suspition");
        }
      }

      sendFail(new_failed,ev.getChannel());
    }
  }

  private void handleSuspectTimer(SuspectTimer ev) {
    int i;
    boolean new_fail=false;
    boolean[] new_failed=null;

    for (i=0 ; i < suspects.length ; i++) {
      if (i != ls.my_rank) {
        suspects[i]++;
        if ( (suspects[i] > suspect_idle) && !ls.failed[i] ) {
          ls.fail(i);
          if (!new_fail) {
            new_fail=true;
            new_failed=new boolean[ls.failed.length];
          }
          new_failed[i]=true;
        }
      }
    }

    if (new_fail) {
      try {
        Suspect suspect=new Suspect(ls.failed,ev.getChannel(),new Direction(Direction.DOWN),this,vs.group,vs.id);
        suspect.getObjectsMessage().push(ls.failed);
        suspect.go();
      } catch (AppiaEventException ex) {
        ex.printStackTrace();
        System.err.println("appia:group:SuspectSession: impossible to propagat local suspition");
      }

      sendFail(new_failed,ev.getChannel());
    }

    try {
      Alive alive=new Alive(ev.getChannel(),new Direction(Direction.DOWN),this,vs.group,vs.id);
      alive.go();
    } catch (AppiaEventException ex) {
      ex.printStackTrace();
      System.err.println("appia:group:SuspectSession: impossible to send ping");
    }
  }

  private void sendFail(boolean[] failed, Channel channel) {
    try {
      Fail ev=new Fail(failed,vs.group,vs.id);
      EchoEvent echo=new EchoEvent(ev,channel,new Direction(Direction.UP),this);
      echo.go();
    } catch (AppiaEventException ex) {
      ex.printStackTrace();
      System.err.println("appia:group:SuspectSession: impossible to inform locally of failure");
    }
  }
}