package appia.protocols.udpsimple;

//////////////////////////////////////////////////////////////////////
//                                                                  //
// Apia: protocol development and composition framework             //
//                                                                  //
// Version: 1.0/J                                                   //
//                                                                  //
// Copyright, 2000, Universidade de Lisboa                          //
// All rights reserved                                              //
// See license.txt for further information                          //
//                                                                  //
// Class: UdpSimpleReader: Thread polling datagrams from opened     //
//                         sockets                                  //
//                                                                  //
// Author: Hugo Miranda, 05/2000                                    //
//                                                                  //
// Change Log:                                                      //
// Data, alteracao, razao/consequncias                             //
//                                                                  //
//////////////////////////////////////////////////////////////////////

import appia.*;
import appia.events.*;
import java.net.*;

/**
 * This is the class responsible for blocking on a socket waiting for
 * datagrams to come. Incoming datagrams are transformed in Events and 
 * wait that the main thread pools them.
 *
 * @see apia.Session
 * @author Hugo Miranda
 */

class UdpSimpleReader extends Thread {

//    private final boolean debugOn=false;

    private DatagramSocket      s;
    private UdpSimpleSession    parentSession;
    public  SendableEvent       waitingEvent;
    public  java.io.PrintStream debugOutput=null;

    public UdpSimpleReader(UdpSimpleSession parentSession, 
			   DatagramSocket s) {
	super();
	this.s=s;
	this.parentSession=parentSession;
	waitingEvent=null;
    }

    public void run() {
	while(true) {
	    try {
		int size=s.getReceiveBufferSize();
		byte[] b=new byte[size];
		DatagramPacket msg = new DatagramPacket(b,b.length);
		s.receive(msg);
/*		if(debugOn && debugOutput!=null)
		    debugOutput.println("UDP: Reader received packet");
*/		SendableEvent e=receiveAndFormat(msg);
                if(e!=null) {
		    /* New event expecting to be sent. Requests a
		       asynchronous event.
		    */
		    synchronized(this) {
			e.getChannel().makeAsync();
/*			if(debugOn && debugOutput!=null)
			    debugOutput.println("UDP: Reader async requested");
*/			waitingEvent=e;
			while(waitingEvent!=null) {
			    wait();
			}
		    }
		}
	    }
	    catch(java.io.IOException e) {}
	    catch(InterruptedException e) {}
	    catch(AppiaEventException e) {}
	}
    }

    /** Event deserialization. Returns the event or null if something
     * happened.
     */

    private SendableEvent receiveAndFormat(DatagramPacket p) {
	byte[] data=p.getData();
	int    curPos=0;
	SendableEvent e=null;

	/* Extract event class name */
	int sLength=data[curPos];
	String className=new String(data,curPos+1,sLength);
	curPos+=data[curPos]+1;

	/* Create event */

	try {
	    Class c = Class.forName(className);
/*	    if(debugOn && debugOutput!=null)
		debugOutput.println("UDP: Datagram received. Creating "+
				    "an "+ className + " event.");
*/	    e=(SendableEvent)c.newInstance();

	    /* Format known event attributes */

	    e.setDirection(new Direction(Direction.UP));
	    e.setSource(parentSession);
	
	    /* Extract channel name and put event in it*/
	    
	    String channelName=new String(data,curPos+1,data[curPos]);
	    Channel msgChannel=findChannel(channelName);

	    /* If channel does not exist, discard message */
	    if(msgChannel==null) return null;

	    e.setChannel(msgChannel);
	    curPos+=data[curPos]+1;

	    /* Extract the addresses and put them on the event */

	    String sourceName=new String(data,curPos+1,data[curPos]);
	    e.source=new InetWithPort();
	    ((InetWithPort)e.source).host=InetAddress.getByName(sourceName);
	    ((InetWithPort)e.source).port=p.getPort();
	    curPos+=data[curPos]+1;

	    String destName=new String(data,curPos+1,data[curPos]);
	    e.dest=new InetWithPort();
	    ((InetWithPort)e.dest).host=InetAddress.getByName(destName);
	    ((InetWithPort)e.dest).port=s.getPort();
	    curPos+=data[curPos]+1;
	    
	    /* Finally, the message */
	    e.getMessage().setByteArray(data,curPos,p.getLength()-curPos);
	}
	catch(ClassNotFoundException ex) {
	    /* The event class is not defined at this endpoint. Give up */
	    System.err.println("Could not raise event due to a class not "+
			       "found exception");
	}
	catch(InstantiationException ex) {
	    /* An instance could not be created. Give up */
	    System.err.println("Could not raise event due to an "+
			       "instantiation exception");
	}
	catch(IllegalAccessException ex) {
	    /* An instance could not be created. Give up */
	    System.err.println("Could not raise event due to an "+
			       "illegal access exception");
	}
	catch(UnknownHostException ex) {
	    /* Host addresses are wrong */
	    System.err.println("Unknown Host Exception in UdpSimpleSession");
	}
	return e;
    }

    /* This method locates an existing channel with the specified
       channelName. If none is found returns null.
    */
    private Channel findChannel(String channelName) {
	java.util.Enumeration e     = parentSession.channels.elements();
	          Channel     found = null;

	while(found==null && e.hasMoreElements()) {
	    Channel c=(Channel)(e.nextElement());
	    if(channelName.equals(c.getChannelID())) {
		found=c;
	    }
	}
	return found;
    }

}
